@@ -69,10 +69,12 @@ use self::ConfigValue as CV;
 use crate::core::compiler::rustdoc::RustdocExternMap;
 use crate::core::shell::Verbosity;
 use crate::core::{features, CliUnstable, Shell, SourceId, Workspace, WorkspaceRootConfig};
-use crate::ops;
+use crate::ops::{self, RegistryCredentialConfig};
+use crate::util::auth::Secret;
 use crate::util::errors::CargoResult;
-use crate::util::toml as cargo_toml;
 use crate::util::validate_package_name;
+use crate::util::CanonicalUrl;
+use crate::util::{internal, toml as cargo_toml};
 use crate::util::{FileLock, Filesystem, IntoUrl, IntoUrlWithBase, Rustc};
 use anyhow::{anyhow, bail, format_err, Context as _};
 use cargo_util::paths;
@@ -135,6 +137,18 @@ enum WhyLoad {
     FileDiscovery,
 }
 
+/// A previously generated authentication token and the data needed to determine if it can be reused.
+#[derive(Debug)]
+pub struct CredentialCacheValue {
+    /// If the command line was used to override the token then it must always be reused,
+    /// even if reading the configuration files would lead to a different value.
+    pub from_commandline: bool,
+    /// If nothing depends on which endpoint is being hit, then we can reuse the token
+    /// for any future request even if some of the requests involve mutations.
+    pub independent_of_endpoint: bool,
+    pub token_value: Secret<String>,
+}
+
 /// Configuration information for cargo. This is not specific to a build, it is information
 /// relating to cargo itself.
 #[derive(Debug)]
@@ -145,6 +159,8 @@ pub struct Config {
     shell: RefCell<Shell>,
     /// A collection of configuration options
     values: LazyCell<HashMap<String, ConfigValue>>,
+    /// A collection of configuration options from the credentials file
+    credential_values: LazyCell<HashMap<String, ConfigValue>>,
     /// CLI config values, passed in via `configure`.
     cli_config: Option<Vec<String>>,
     /// The current working directory of cargo
@@ -188,6 +204,9 @@ pub struct Config {
     upper_case_env: HashMap<String, String>,
     /// Tracks which sources have been updated to avoid multiple updates.
     updated_sources: LazyCell<RefCell<HashSet<SourceId>>>,
+    /// Cache of credentials from configuration or credential providers.
+    /// Maps from url to credential value.
+    credential_cache: LazyCell<RefCell<HashMap<CanonicalUrl, CredentialCacheValue>>>,
     /// Lock, if held, of the global package cache along with the number of
     /// acquisitions so far.
     package_cache_lock: RefCell<Option<(Option<FileLock>, usize)>>,
@@ -267,6 +286,7 @@ impl Config {
             cwd,
             search_stop_path: None,
             values: LazyCell::new(),
+            credential_values: LazyCell::new(),
             cli_config: None,
             cargo_exe: LazyCell::new(),
             rustdoc: LazyCell::new(),
@@ -291,6 +311,7 @@ impl Config {
             env,
             upper_case_env,
             updated_sources: LazyCell::new(),
+            credential_cache: LazyCell::new(),
             package_cache_lock: RefCell::new(None),
             http_config: LazyCell::new(),
             future_incompat_config: LazyCell::new(),
@@ -327,6 +348,18 @@ impl Config {
         &self.home_path
     }
 
+    /// Returns a path to display to the user with the location of their home
+    /// config file (to only be used for displaying a diagnostics suggestion,
+    /// such as recommending where to add a config value).
+    pub fn diagnostic_home_config(&self) -> String {
+        let home = self.home_path.as_path_unlocked();
+        let path = match self.get_file_path(home, "config", false) {
+            Ok(Some(existing_path)) => existing_path,
+            _ => home.join("config.toml"),
+        };
+        path.to_string_lossy().to_string()
+    }
+
     /// Gets the Cargo Git directory (`<cargo_home>/git`).
     pub fn git_path(&self) -> Filesystem {
         self.home_path.join("git")
@@ -406,6 +439,18 @@ impl Config {
     pub fn cargo_exe(&self) -> CargoResult<&Path> {
         self.cargo_exe
             .try_borrow_with(|| {
+                fn from_env() -> CargoResult<PathBuf> {
+                    // Try re-using the `cargo` set in the environment already. This allows
+                    // commands that use Cargo as a library to inherit (via `cargo <subcommand>`)
+                    // or set (by setting `$CARGO`) a correct path to `cargo` when the current exe
+                    // is not actually cargo (e.g., `cargo-*` binaries, Valgrind, `ld.so`, etc.).
+                    let exe = env::var_os(crate::CARGO_ENV)
+                        .map(PathBuf::from)
+                        .ok_or_else(|| anyhow!("$CARGO not set"))?
+                        .canonicalize()?;
+                    Ok(exe)
+                }
+
                 fn from_current_exe() -> CargoResult<PathBuf> {
                     // Try fetching the path to `cargo` using `env::current_exe()`.
                     // The method varies per operating system and might fail; in particular,
@@ -431,7 +476,8 @@ impl Config {
                     paths::resolve_executable(&argv0)
                 }
 
-                let exe = from_current_exe()
+                let exe = from_env()
+                    .or_else(|_| from_current_exe())
                     .or_else(|_| from_argv())
                     .with_context(|| "couldn't get the path to cargo executable")?;
                 Ok(exe)
@@ -446,6 +492,13 @@ impl Config {
             .borrow_mut()
     }
 
+    /// Cached credentials from credential providers or configuration.
+    pub fn credential_cache(&self) -> RefMut<'_, HashMap<CanonicalUrl, CredentialCacheValue>> {
+        self.credential_cache
+            .borrow_with(|| RefCell::new(HashMap::new()))
+            .borrow_mut()
+    }
+
     /// Gets all config values from disk.
     ///
     /// This will lazy-load the values as necessary. Callers are responsible
@@ -462,10 +515,11 @@ impl Config {
     /// entries. This doesn't respect environment variables. You should avoid
     /// using this if possible.
     pub fn values_mut(&mut self) -> CargoResult<&mut HashMap<String, ConfigValue>> {
-        match self.values.borrow_mut() {
-            Some(map) => Ok(map),
-            None => bail!("config values not loaded yet"),
-        }
+        let _ = self.values()?;
+        Ok(self
+            .values
+            .borrow_mut()
+            .expect("already loaded config values"))
     }
 
     // Note: this is used by RLS, not Cargo.
@@ -542,8 +596,21 @@ impl Config {
     /// This does NOT look at environment variables. See `get_cv_with_env` for
     /// a variant that supports environment variables.
     fn get_cv(&self, key: &ConfigKey) -> CargoResult<Option<ConfigValue>> {
+        if let Some(vals) = self.credential_values.borrow() {
+            let val = self.get_cv_helper(key, vals)?;
+            if val.is_some() {
+                return Ok(val);
+            }
+        }
+        self.get_cv_helper(key, self.values()?)
+    }
+
+    fn get_cv_helper(
+        &self,
+        key: &ConfigKey,
+        vals: &HashMap<String, ConfigValue>,
+    ) -> CargoResult<Option<ConfigValue>> {
         log::trace!("get cv {:?}", key);
-        let vals = self.values()?;
         if key.is_root() {
             // Returning the entire root table (for example `cargo config get`
             // with no key). The definition here shouldn't matter.
@@ -1318,6 +1385,26 @@ impl Config {
                     );
                 }
 
+                if toml_v
+                    .get("registry")
+                    .and_then(|v| v.as_table())
+                    .and_then(|t| t.get("secret-key"))
+                    .is_some()
+                {
+                    bail!(
+                        "registry.secret-key cannot be set through --config for security reasons"
+                    );
+                } else if let Some((k, _)) = toml_v
+                    .get("registries")
+                    .and_then(|v| v.as_table())
+                    .and_then(|t| t.iter().find(|(_, v)| v.get("secret-key").is_some()))
+                {
+                    bail!(
+                        "registries.{}.secret-key cannot be set through --config for security reasons",
+                        k
+                    );
+                }
+
                 CV::from_toml(Definition::Cli(None), toml_v)
                     .with_context(|| format!("failed to convert --config argument `{arg}`"))?
             };
@@ -1337,8 +1424,6 @@ impl Config {
             CV::Table(table, _def) => table,
             _ => unreachable!(),
         };
-        // Force values to be loaded.
-        let _ = self.values()?;
         let values = self.values_mut()?;
         for (key, value) in loaded_map.into_iter() {
             match values.entry(key) {
@@ -1469,7 +1554,17 @@ impl Config {
     }
 
     /// Loads credentials config from the credentials file, if present.
-    pub fn load_credentials(&mut self) -> CargoResult<()> {
+    ///
+    /// The credentials are loaded into a separate field to enable them
+    /// to be lazy-loaded after the main configuration has been loaded,
+    /// without requiring `mut` access to the `Config`.
+    ///
+    /// If the credentials are already loaded, this function does nothing.
+    pub fn load_credentials(&self) -> CargoResult<()> {
+        if self.credential_values.filled() {
+            return Ok(());
+        }
+
         let home_path = self.home_path.clone().into_path_unlocked();
         let credentials = match self.get_file_path(&home_path, "credentials", true)? {
             Some(credentials) => credentials,
@@ -1493,20 +1588,24 @@ impl Config {
             }
         }
 
+        let mut credential_values = HashMap::new();
         if let CV::Table(map, _) = value {
-            let base_map = self.values_mut()?;
+            let base_map = self.values()?;
             for (k, v) in map {
-                match base_map.entry(k) {
-                    Vacant(entry) => {
-                        entry.insert(v);
+                let entry = match base_map.get(&k) {
+                    Some(base_entry) => {
+                        let mut entry = base_entry.clone();
+                        entry.merge(v, true)?;
+                        entry
                     }
-                    Occupied(mut entry) => {
-                        entry.get_mut().merge(v, true)?;
-                    }
-                }
+                    None => v,
+                };
+                credential_values.insert(k, entry);
             }
         }
-
+        self.credential_values
+            .fill(credential_values)
+            .expect("was not filled at beginning of the function");
         Ok(())
     }
 
@@ -2027,19 +2126,28 @@ pub fn homedir(cwd: &Path) -> Option<PathBuf> {
 
 pub fn save_credentials(
     cfg: &Config,
-    token: Option<String>,
-    registry: Option<&str>,
+    token: Option<RegistryCredentialConfig>,
+    registry: &SourceId,
 ) -> CargoResult<()> {
-    // If 'credentials.toml' exists, we should write to that, otherwise
-    // use the legacy 'credentials'. There's no need to print the warning
-    // here, because it would already be printed at load time.
+    let registry = if registry.is_crates_io() {
+        None
+    } else {
+        let name = registry
+            .alt_registry_key()
+            .ok_or_else(|| internal("can't save credentials for anonymous registry"))?;
+        Some(name)
+    };
+
+    // If 'credentials' exists, write to that for backward compatibility reasons.
+    // Otherwise write to 'credentials.toml'. There's no need to print the
+    // warning here, because it would already be printed at load time.
     let home_path = cfg.home_path.clone().into_path_unlocked();
     let filename = match cfg.get_file_path(&home_path, "credentials", false)? {
         Some(path) => match path.file_name() {
             Some(filename) => Path::new(filename).to_owned(),
-            None => Path::new("credentials").to_owned(),
+            None => Path::new("credentials.toml").to_owned(),
         },
-        None => Path::new("credentials").to_owned(),
+        None => Path::new("credentials.toml").to_owned(),
     };
 
     let mut file = {
@@ -2068,26 +2176,50 @@ pub fn save_credentials(
 
     if let Some(token) = token {
         // login
-        let (key, mut value) = {
-            let key = "token".to_string();
-            let value = ConfigValue::String(token, Definition::Path(file.path().to_path_buf()));
-            let map = HashMap::from([(key, value)]);
-            let table = CV::Table(map, Definition::Path(file.path().to_path_buf()));
 
-            if let Some(registry) = registry {
-                let map = HashMap::from([(registry.to_string(), table)]);
-                (
-                    "registries".into(),
-                    CV::Table(map, Definition::Path(file.path().to_path_buf())),
-                )
-            } else {
-                ("registry".into(), table)
+        let path_def = Definition::Path(file.path().to_path_buf());
+        let (key, mut value) = match token {
+            RegistryCredentialConfig::Token(token) => {
+                // login with token
+
+                let key = "token".to_string();
+                let value = ConfigValue::String(token.expose(), path_def.clone());
+                let map = HashMap::from([(key, value)]);
+                let table = CV::Table(map, path_def.clone());
+
+                if let Some(registry) = registry {
+                    let map = HashMap::from([(registry.to_string(), table)]);
+                    ("registries".into(), CV::Table(map, path_def.clone()))
+                } else {
+                    ("registry".into(), table)
+                }
             }
+            RegistryCredentialConfig::AsymmetricKey((secret_key, key_subject)) => {
+                // login with key
+
+                let key = "secret-key".to_string();
+                let value = ConfigValue::String(secret_key.expose(), path_def.clone());
+                let mut map = HashMap::from([(key, value)]);
+                if let Some(key_subject) = key_subject {
+                    let key = "secret-key-subject".to_string();
+                    let value = ConfigValue::String(key_subject, path_def.clone());
+                    map.insert(key, value);
+                }
+                let table = CV::Table(map, path_def.clone());
+
+                if let Some(registry) = registry {
+                    let map = HashMap::from([(registry.to_string(), table)]);
+                    ("registries".into(), CV::Table(map, path_def.clone()))
+                } else {
+                    ("registry".into(), table)
+                }
+            }
+            _ => unreachable!(),
         };
 
         if registry.is_some() {
             if let Some(table) = toml.as_table_mut().unwrap().remove("registries") {
-                let v = CV::from_toml(Definition::Path(file.path().to_path_buf()), table)?;
+                let v = CV::from_toml(path_def, table)?;
                 value.merge(v, false)?;
             }
         }
@@ -2102,6 +2234,8 @@ pub fn save_credentials(
                         format_err!("expected `[registries.{}]` to be a table", registry)
                     })?;
                     rtable.remove("token");
+                    rtable.remove("secret-key");
+                    rtable.remove("secret-key-subject");
                 }
             }
         } else if let Some(registry) = table.get_mut("registry") {
@@ -2109,6 +2243,8 @@ pub fn save_credentials(
                 .as_table_mut()
                 .ok_or_else(|| format_err!("expected `[registry]` to be a table"))?;
             reg_table.remove("token");
+            reg_table.remove("secret-key");
+            reg_table.remove("secret-key-subject");
         }
     }
 
@@ -2224,6 +2360,13 @@ pub struct CargoNetConfig {
     pub retry: Option<u32>,
     pub offline: Option<bool>,
     pub git_fetch_with_cli: Option<bool>,
+    pub ssh: Option<CargoSshConfig>,
+}
+
+#[derive(Debug, Deserialize)]
+#[serde(rename_all = "kebab-case")]
+pub struct CargoSshConfig {
+    pub known_hosts: Option<Vec<Value<String>>>,
 }
 
 #[derive(Debug, Deserialize)]
