@@ -1,8 +1,8 @@
 //! Tests for credential-process.
 
+use cargo_test_support::registry::{Package, TestRegistry};
 use cargo_test_support::{basic_manifest, cargo_process, paths, project, registry, Project};
-use std::fs;
-use std::thread;
+use std::fs::{self, read_to_string};
 
 fn toml_bin(proj: &Project, name: &str) -> String {
     proj.bin(name).display().to_string().replace('\\', "\\\\")
@@ -10,9 +10,13 @@ fn toml_bin(proj: &Project, name: &str) -> String {
 
 #[cargo_test]
 fn gated() {
-    registry::RegistryBuilder::new()
-        .alternative(true)
-        .add_tokens(false)
+    let _alternative = registry::RegistryBuilder::new()
+        .alternative()
+        .no_configure_token()
+        .build();
+
+    let cratesio = registry::RegistryBuilder::new()
+        .no_configure_token()
         .build();
 
     let p = project()
@@ -28,12 +32,14 @@ fn gated() {
         .build();
 
     p.cargo("publish --no-verify")
-        .masquerade_as_nightly_cargo()
+        .replace_crates_io(cratesio.index_url())
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_status(101)
         .with_stderr(
             "\
 [UPDATING] [..]
-[ERROR] no upload token found, please run `cargo login` or pass `--token`
+[ERROR] no token found, please run `cargo login`
+or use environment variable CARGO_REGISTRY_TOKEN
 ",
         )
         .run();
@@ -47,12 +53,13 @@ fn gated() {
     );
 
     p.cargo("publish --no-verify --registry alternative")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_status(101)
         .with_stderr(
             "\
 [UPDATING] [..]
-[ERROR] no upload token found, please run `cargo login` or pass `--token`
+[ERROR] no token found for `alternative`, please run `cargo login --registry alternative`
+or use environment variable CARGO_REGISTRIES_ALTERNATIVE_TOKEN
 ",
         )
         .run();
@@ -61,9 +68,9 @@ fn gated() {
 #[cargo_test]
 fn warn_both_token_and_process() {
     // Specifying both credential-process and a token in config should issue a warning.
-    registry::RegistryBuilder::new()
-        .alternative(true)
-        .add_tokens(false)
+    let _server = registry::RegistryBuilder::new()
+        .alternative()
+        .no_configure_token()
         .build();
     let p = project()
         .file(
@@ -89,13 +96,23 @@ fn warn_both_token_and_process() {
         .file("src/lib.rs", "")
         .build();
 
+    // HACK: Inject `foo` directly into the index so `publish` won't block for it to be in
+    // the index.
+    //
+    // This is to ensure we can verify the Summary we post to the registry as doing so precludes
+    // the registry from processing the publish.
+    Package::new("foo", "0.1.0")
+        .file("src/lib.rs", "")
+        .alternative(true)
+        .publish();
+
     p.cargo("publish --no-verify --registry alternative -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_status(101)
         .with_stderr(
             "\
-[ERROR] both `registries.alternative.token` and `registries.alternative.credential-process` \
-were specified in the config\n\
+[UPDATING] [..]
+[ERROR] both `token` and `credential-process` were specified in the config for registry `alternative`.
 Only one of these values may be set, remove one or the other to proceed.
 ",
         )
@@ -114,12 +131,14 @@ Only one of these values may be set, remove one or the other to proceed.
         "#,
     );
     p.cargo("publish --no-verify --registry alternative -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_stderr(
             "\
 [UPDATING] [..]
 [PACKAGING] foo v0.1.0 [..]
+[PACKAGED] [..]
 [UPLOADING] foo v0.1.0 [..]
+[UPDATING] [..]
 ",
         )
         .run();
@@ -134,25 +153,37 @@ Only one of these values may be set, remove one or the other to proceed.
 /// * Create a simple `foo` project to run the test against.
 /// * Configure the credential-process config.
 ///
-/// Returns a thread handle for the API server, the test should join it when
-/// finished. Also returns the simple `foo` project to test against.
-fn get_token_test() -> (Project, thread::JoinHandle<()>) {
+/// Returns the simple `foo` project to test against and the API server handle.
+fn get_token_test() -> (Project, TestRegistry) {
     // API server that checks that the token is included correctly.
     let server = registry::RegistryBuilder::new()
-        .add_tokens(false)
-        .build_api_server(&|headers| {
-            assert!(headers
-                .iter()
-                .any(|header| header == "Authorization: sekrit"));
-
-            (200, &r#"{"ok": true, "msg": "completed!"}"#)
-        });
-
+        .no_configure_token()
+        .token(cargo_test_support::registry::Token::Plaintext(
+            "sekrit".to_string(),
+        ))
+        .alternative()
+        .http_api()
+        .build();
     // The credential process to use.
     let cred_proj = project()
         .at("cred_proj")
         .file("Cargo.toml", &basic_manifest("test-cred", "1.0.0"))
-        .file("src/main.rs", r#"fn main() { println!("sekrit"); } "#)
+        .file(
+            "src/main.rs",
+            r#"
+                use std::fs::File;
+                use std::io::Write;
+                fn main() {
+                    let mut f = File::options()
+                        .write(true)
+                        .create(true)
+                        .append(true)
+                        .open("runs.log")
+                        .unwrap();
+                    write!(f, "+");
+                    println!("sekrit");
+                } "#,
+        )
         .build();
     cred_proj.cargo("build").run();
 
@@ -165,7 +196,7 @@ fn get_token_test() -> (Project, thread::JoinHandle<()>) {
                     index = "{}"
                     credential-process = ["{}"]
                 "#,
-                registry::alt_registry_url(),
+                server.index_url(),
                 toml_bin(&cred_proj, "test-cred")
             ),
         )
@@ -189,26 +220,31 @@ fn get_token_test() -> (Project, thread::JoinHandle<()>) {
 #[cargo_test]
 fn publish() {
     // Checks that credential-process is used for `cargo publish`.
-    let (p, t) = get_token_test();
+    let (p, _t) = get_token_test();
 
     p.cargo("publish --no-verify --registry alternative -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_stderr(
             "\
 [UPDATING] [..]
 [PACKAGING] foo v0.1.0 [..]
+[PACKAGED] [..]
 [UPLOADING] foo v0.1.0 [..]
+[UPDATING] [..]
 ",
         )
         .run();
 
-    t.join().ok().unwrap();
+    let calls = read_to_string(p.root().join("runs.log")).unwrap().len();
+    assert_eq!(calls, 1);
 }
 
 #[cargo_test]
 fn basic_unsupported() {
     // Non-action commands don't support login/logout.
-    registry::RegistryBuilder::new().add_tokens(false).build();
+    let registry = registry::RegistryBuilder::new()
+        .no_configure_token()
+        .build();
     cargo_util::paths::append(
         &paths::home().join(".cargo/config"),
         br#"
@@ -219,11 +255,12 @@ fn basic_unsupported() {
     .unwrap();
 
     cargo_process("login -Z credential-process abcdefg")
-        .masquerade_as_nightly_cargo()
+        .replace_crates_io(registry.index_url())
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_status(101)
         .with_stderr(
             "\
-[UPDATING] [..]
+[UPDATING] crates.io index
 [ERROR] credential process `false` cannot be used to log in, \
 the credential-process configuration value must pass the \
 `{action}` argument in the config to support this command
@@ -232,7 +269,8 @@ the credential-process configuration value must pass the \
         .run();
 
     cargo_process("logout -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .replace_crates_io(registry.index_url())
+        .masquerade_as_nightly_cargo(&["credential-process", "cargo-logout"])
         .with_status(101)
         .with_stderr(
             "\
@@ -246,27 +284,28 @@ the credential-process configuration value must pass the \
 
 #[cargo_test]
 fn login() {
-    registry::init();
+    let server = registry::RegistryBuilder::new()
+        .no_configure_token()
+        .build();
     // The credential process to use.
     let cred_proj = project()
         .at("cred_proj")
         .file("Cargo.toml", &basic_manifest("test-cred", "1.0.0"))
         .file(
             "src/main.rs",
-            &r#"
+                r#"
                 use std::io::Read;
 
-                fn main() {
-                    assert_eq!(std::env::var("CARGO_REGISTRY_NAME").unwrap(), "crates-io");
-                    assert_eq!(std::env::var("CARGO_REGISTRY_API_URL").unwrap(), "__API__");
+                fn main() {{
+                    assert_eq!(std::env::var("CARGO_REGISTRY_NAME_OPT").unwrap(), "crates-io");
+                    assert_eq!(std::env::var("CARGO_REGISTRY_INDEX_URL").unwrap(), "https://github.com/rust-lang/crates.io-index");
                     assert_eq!(std::env::args().skip(1).next().unwrap(), "store");
                     let mut buffer = String::new();
                     std::io::stdin().read_to_string(&mut buffer).unwrap();
                     assert_eq!(buffer, "abcdefg\n");
                     std::fs::write("token-store", buffer).unwrap();
-                }
-            "#
-            .replace("__API__", &registry::api_url().to_string()),
+                }}
+            "#,
         )
         .build();
     cred_proj.cargo("build").run();
@@ -285,7 +324,8 @@ fn login() {
     .unwrap();
 
     cargo_process("login -Z credential-process abcdefg")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
+        .replace_crates_io(server.index_url())
         .with_stderr(
             "\
 [UPDATING] [..]
@@ -301,23 +341,25 @@ fn login() {
 
 #[cargo_test]
 fn logout() {
-    registry::RegistryBuilder::new().add_tokens(false).build();
+    let server = registry::RegistryBuilder::new()
+        .no_configure_token()
+        .build();
     // The credential process to use.
     let cred_proj = project()
         .at("cred_proj")
         .file("Cargo.toml", &basic_manifest("test-cred", "1.0.0"))
         .file(
             "src/main.rs",
-            r#"
+                r#"
                 use std::io::Read;
 
-                fn main() {
-                    assert_eq!(std::env::var("CARGO_REGISTRY_NAME").unwrap(), "crates-io");
+                fn main() {{
+                    assert_eq!(std::env::var("CARGO_REGISTRY_NAME_OPT").unwrap(), "crates-io");
+                    assert_eq!(std::env::var("CARGO_REGISTRY_INDEX_URL").unwrap(), "https://github.com/rust-lang/crates.io-index");
                     assert_eq!(std::env::args().skip(1).next().unwrap(), "erase");
                     std::fs::write("token-store", "").unwrap();
-                    eprintln!("token for `{}` has been erased!",
-                        std::env::var("CARGO_REGISTRY_NAME").unwrap());
-                }
+                    eprintln!("token for `crates-io` has been erased!")
+                }}
             "#,
         )
         .build();
@@ -337,12 +379,12 @@ fn logout() {
     .unwrap();
 
     cargo_process("logout -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process", "cargo-logout"])
+        .replace_crates_io(server.index_url())
         .with_stderr(
             "\
-[UPDATING] [..]
 token for `crates-io` has been erased!
-[LOGOUT] token for `crates.io` has been removed from local storage
+[LOGOUT] token for `crates-io` has been removed from local storage
 ",
         )
         .run();
@@ -354,10 +396,10 @@ token for `crates-io` has been erased!
 
 #[cargo_test]
 fn yank() {
-    let (p, t) = get_token_test();
+    let (p, _t) = get_token_test();
 
     p.cargo("yank --version 0.1.0 --registry alternative -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_stderr(
             "\
 [UPDATING] [..]
@@ -365,16 +407,14 @@ fn yank() {
 ",
         )
         .run();
-
-    t.join().ok().unwrap();
 }
 
 #[cargo_test]
 fn owner() {
-    let (p, t) = get_token_test();
+    let (p, _t) = get_token_test();
 
     p.cargo("owner --add username --registry alternative -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_stderr(
             "\
 [UPDATING] [..]
@@ -382,14 +422,14 @@ fn owner() {
 ",
         )
         .run();
-
-    t.join().ok().unwrap();
 }
 
 #[cargo_test]
 fn libexec_path() {
     // cargo: prefixed names use the sysroot
-    registry::RegistryBuilder::new().add_tokens(false).build();
+    let server = registry::RegistryBuilder::new()
+        .no_configure_token()
+        .build();
     cargo_util::paths::append(
         &paths::home().join(".cargo/config"),
         br#"
@@ -400,7 +440,8 @@ fn libexec_path() {
     .unwrap();
 
     cargo_process("login -Z credential-process abcdefg")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
+        .replace_crates_io(server.index_url())
         .with_status(101)
         .with_stderr(
             // FIXME: Update "Caused by" error message once rust/pull/87704 is merged.
@@ -420,9 +461,9 @@ Caused by:
 #[cargo_test]
 fn invalid_token_output() {
     // Error when credential process does not output the expected format for a token.
-    registry::RegistryBuilder::new()
-        .alternative(true)
-        .add_tokens(false)
+    let _server = registry::RegistryBuilder::new()
+        .alternative()
+        .no_configure_token()
         .build();
     let cred_proj = project()
         .at("cred_proj")
@@ -450,7 +491,7 @@ fn invalid_token_output() {
         .build();
 
     p.cargo("publish --no-verify --registry alternative -Z credential-process")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["credential-process"])
         .with_status(101)
         .with_stderr(
             "\
