@@ -4,7 +4,7 @@ use cargo::core::compiler::RustDocFingerprint;
 use cargo_test_support::paths::CargoPathExt;
 use cargo_test_support::registry::Package;
 use cargo_test_support::{basic_lib_manifest, basic_manifest, git, project};
-use cargo_test_support::{is_nightly, rustc_host, symlink_supported, tools};
+use cargo_test_support::{rustc_host, symlink_supported, tools};
 use std::fs;
 use std::str;
 
@@ -748,12 +748,8 @@ fn doc_same_name() {
     p.cargo("doc").run();
 }
 
-#[cargo_test]
+#[cargo_test(nightly, reason = "no_core, lang_items requires nightly")]
 fn doc_target() {
-    if !is_nightly() {
-        // no_core, lang_items requires nightly.
-        return;
-    }
     const TARGET: &str = "arm-unknown-linux-gnueabihf";
 
     let p = project()
@@ -1114,7 +1110,7 @@ fn doc_all_workspace() {
         .file(
             "Cargo.toml",
             r#"
-                [project]
+                [package]
                 name = "foo"
                 version = "0.1.0"
 
@@ -1250,7 +1246,7 @@ fn doc_all_member_dependency_same_name() {
         .file(
             "bar/Cargo.toml",
             r#"
-                [project]
+                [package]
                 name = "bar"
                 version = "0.1.0"
 
@@ -1310,13 +1306,8 @@ fn doc_workspace_open_help_message() {
         .run();
 }
 
-#[cargo_test]
+#[cargo_test(nightly, reason = "-Zextern-html-root-url is unstable")]
 fn doc_extern_map_local() {
-    if !is_nightly() {
-        // -Zextern-html-root-url is unstable
-        return;
-    }
-
     let p = project()
         .file(
             "Cargo.toml",
@@ -1332,7 +1323,7 @@ fn doc_extern_map_local() {
 
     p.cargo("doc -v --no-deps -Zrustdoc-map --open")
         .env("BROWSER", tools::echo())
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["rustdoc-map"])
         .with_stderr(
             "\
 [DOCUMENTING] foo v0.1.0 [..]
@@ -2050,13 +2041,8 @@ fn crate_versions_flag_is_overridden() {
     asserts(output_documentation());
 }
 
-#[cargo_test]
+#[cargo_test(nightly, reason = "-Zdoctest-in-workspace is unstable")]
 fn doc_test_in_workspace() {
-    if !is_nightly() {
-        // -Zdoctest-in-workspace is unstable
-        return;
-    }
-
     let p = project()
         .file(
             "Cargo.toml",
@@ -2071,7 +2057,7 @@ fn doc_test_in_workspace() {
         .file(
             "crate-a/Cargo.toml",
             r#"
-                [project]
+                [package]
                 name = "crate-a"
                 version = "0.1.0"
             "#,
@@ -2087,7 +2073,7 @@ fn doc_test_in_workspace() {
         .file(
             "crate-b/Cargo.toml",
             r#"
-                [project]
+                [package]
                 name = "crate-b"
                 version = "0.1.0"
             "#,
@@ -2102,7 +2088,7 @@ fn doc_test_in_workspace() {
         )
         .build();
     p.cargo("test -Zdoctest-in-workspace --doc -vv")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["doctest-in-workspace"])
         .with_stderr_contains("[DOCTEST] crate-a")
         .with_stdout_contains(
             "
@@ -2346,7 +2332,7 @@ fn doc_fingerprint_unusual_behavior() {
     p.change_file("src/lib.rs", "// changed2");
     fs::write(real_doc.join("somefile"), "test").unwrap();
     p.cargo("doc -Z skip-rustdoc-fingerprint")
-        .masquerade_as_nightly_cargo()
+        .masquerade_as_nightly_cargo(&["skip-rustdoc-fingerprint"])
         .with_stderr(
             "[DOCUMENTING] foo [..]\n\
              [FINISHED] [..]",
@@ -2357,209 +2343,6 @@ fn doc_fingerprint_unusual_behavior() {
     assert!(real_doc.join("somefile").exists());
 }
 
-#[cargo_test]
-fn scrape_examples_basic() {
-    if !is_nightly() {
-        // -Z rustdoc-scrape-examples is unstable
-        return;
-    }
-
-    let p = project()
-        .file(
-            "Cargo.toml",
-            r#"
-                [package]
-                name = "foo"
-                version = "0.0.1"
-                authors = []
-            "#,
-        )
-        .file("examples/ex.rs", "fn main() { foo::foo(); }")
-        .file("src/lib.rs", "pub fn foo() {}\npub fn bar() { foo(); }")
-        .build();
-
-    p.cargo("doc -Zunstable-options -Z rustdoc-scrape-examples=all")
-        .masquerade_as_nightly_cargo()
-        .with_stderr(
-            "\
-[..] foo v0.0.1 ([CWD])
-[..] foo v0.0.1 ([CWD])
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]
-",
-        )
-        .run();
-
-    let doc_html = p.read_file("target/doc/foo/fn.foo.html");
-    assert!(doc_html.contains("Examples found in repository"));
-    assert!(doc_html.contains("More examples"));
-
-    // Ensure that the reverse-dependency has its sources generated
-    assert!(p.build_dir().join("doc/src/ex/ex.rs.html").exists());
-}
-
-#[cargo_test]
-fn scrape_examples_avoid_build_script_cycle() {
-    if !is_nightly() {
-        // -Z rustdoc-scrape-examples is unstable
-        return;
-    }
-
-    let p = project()
-        // package with build dependency
-        .file(
-            "Cargo.toml",
-            r#"
-                [package]
-                name = "foo"
-                version = "0.0.1"
-                authors = []
-                links = "foo"
-
-                [workspace]
-                members = ["bar"]
-
-                [build-dependencies]
-                bar = {path = "bar"}
-            "#,
-        )
-        .file("src/lib.rs", "")
-        .file("build.rs", "fn main(){}")
-        // dependency
-        .file(
-            "bar/Cargo.toml",
-            r#"
-                [package]
-                name = "bar"
-                version = "0.0.1"
-                authors = []
-                links = "bar"
-            "#,
-        )
-        .file("bar/src/lib.rs", "")
-        .file("bar/build.rs", "fn main(){}")
-        .build();
-
-    p.cargo("doc --all -Zunstable-options -Z rustdoc-scrape-examples=all")
-        .masquerade_as_nightly_cargo()
-        .run();
-}
-
-#[cargo_test]
-fn scrape_examples_complex_reverse_dependencies() {
-    if !is_nightly() {
-        // -Z rustdoc-scrape-examples is unstable
-        return;
-    }
-
-    let p = project()
-        .file(
-            "Cargo.toml",
-            r#"
-                [package]
-                name = "foo"
-                version = "0.0.1"
-                authors = []
-
-                [dev-dependencies]
-                a = {path = "a", features = ["feature"]}
-                b = {path = "b"}
-
-                [workspace]
-                members = ["b"]
-            "#,
-        )
-        .file("src/lib.rs", "")
-        .file("examples/ex.rs", "fn main() { a::f(); }")
-        .file(
-            "a/Cargo.toml",
-            r#"
-                [package]
-                name = "a"
-                version = "0.0.1"
-                authors = []
-
-                [lib]
-                proc-macro = true
-
-                [dependencies]
-                b = {path = "../b"}
-
-                [features]
-                feature = []
-            "#,
-        )
-        .file("a/src/lib.rs", "#[cfg(feature)] pub fn f();")
-        .file(
-            "b/Cargo.toml",
-            r#"
-                [package]
-                name = "b"
-                version = "0.0.1"
-                authors = []
-            "#,
-        )
-        .file("b/src/lib.rs", "")
-        .build();
-
-    p.cargo("doc -Zunstable-options -Z rustdoc-scrape-examples=all")
-        .masquerade_as_nightly_cargo()
-        .run();
-}
-
-#[cargo_test]
-fn scrape_examples_crate_with_dash() {
-    if !is_nightly() {
-        // -Z rustdoc-scrape-examples is unstable
-        return;
-    }
-
-    let p = project()
-        .file(
-            "Cargo.toml",
-            r#"
-                [package]
-                name = "da-sh"
-                version = "0.0.1"
-                authors = []
-            "#,
-        )
-        .file("src/lib.rs", "pub fn foo() {}")
-        .file("examples/a.rs", "fn main() { da_sh::foo(); }")
-        .build();
-
-    p.cargo("doc -Zunstable-options -Z rustdoc-scrape-examples=all")
-        .masquerade_as_nightly_cargo()
-        .run();
-
-    let doc_html = p.read_file("target/doc/da_sh/fn.foo.html");
-    assert!(doc_html.contains("Examples found in repository"));
-}
-
-#[cargo_test]
-fn scrape_examples_missing_flag() {
-    if !is_nightly() {
-        return;
-    }
-
-    let p = project()
-        .file(
-            "Cargo.toml",
-            r#"
-            [package]
-            name = "foo"
-            version = "1.2.4"
-            authors = []
-        "#,
-        )
-        .file("src/lib.rs", "//! These are the docs!")
-        .build();
-    p.cargo("doc -Zrustdoc-scrape-examples")
-        .masquerade_as_nightly_cargo()
-        .with_status(101)
-        .with_stderr("error: -Z rustdoc-scrape-examples must take [..] an argument")
-        .run();
-}
-
 #[cargo_test]
 fn lib_before_bin() {
     // Checks that the library is documented before the binary.
@@ -2697,3 +2480,24 @@ fn doc_lib_false_dep() {
     assert!(p.build_dir().join("doc/foo").exists());
     assert!(!p.build_dir().join("doc/bar").exists());
 }
+
+#[cargo_test]
+fn link_to_private_item() {
+    let main = r#"
+    //! [bar]
+    #[allow(dead_code)]
+    fn bar() {}
+    "#;
+    let p = project().file("src/lib.rs", main).build();
+    p.cargo("doc")
+        .with_stderr_contains("[..] documentation for `foo` links to private item `bar`")
+        .run();
+    // Check that binaries don't emit a private_intra_doc_links warning.
+    fs::rename(p.root().join("src/lib.rs"), p.root().join("src/main.rs")).unwrap();
+    p.cargo("doc")
+        .with_stderr(
+            "[DOCUMENTING] foo [..]\n\
+             [FINISHED] [..]",
+        )
+        .run();
+}
